"""
A simple bot that favs and reblogs Mastodon statuses which contain a given tag.

The simplest way to use it is to run a cron job that executes:

    `pipenv run python src/mastoboost.py BASE_URL TOKEN USERNAME HASHTAG`

Where

    BASE_URL is the base URL of the Mastodon instance of the account used by the bot
    TOKEN is the application token provided by the account used by the bot
    USERNAME is the username of the account used by the bot
    HASHTAG is the hashtag (without the '#' sign) which the bot must watch
    HOURS determines how far back in time the bot will look for statuses

Options are:

    --hours H: stop fetching when a status is met which is more that H hours old
    --lang LL: select only statuses written in the LL language (LL=en, fr, etc.).
      Example: --lang en
      Can be repeated, e.g. --lang en --lang fr
    --lang-exclude: instead of requesting languages, --lang excludes exclude them.
    --guess-lang: try to guess the status language from its content rather than trust
      its language attribute.
    --reverse: actually unfavourite and unreblog (do not care about language then).
      (otherwise favourite and reblog if not already done)
    --dry-run: do not actually (un)favourite or (un)reblog
    --log-level: set the logging level (DEBUG, INFO, WARNING, etc)
      An additional level is provided: NOTICE, which is between INFO and DEBUG
"""

import logging
from datetime import datetime
import click
from mastodon import Mastodon  # type: ignore
from pytz import utc
from lingua import Language, LanguageDetectorBuilder
from html2text import html2text


def _logging_setup(log_level_name):
    """
    Set up logging and return the new NOTICE log level.
    """
    # Add the "NOTICE" level in between INFO and DEBUG, as we want a level that is more
    # verbose than INFO but will not let other modules display their own logs.
    log_level_notice = (logging.INFO + logging.DEBUG) // 2
    logging.addLevelName(log_level_notice, "NOTICE")
    # See which level we were asked.
    # Yes, despite the name, getLevelName actually also converts string to int!
    log_level_int = logging.getLevelName(log_level_name)
    # Customize the logging format and level.
    logging.basicConfig(format="%(levelname)s: %(message)s", level=log_level_int)
    if log_level_int != logging.INFO:
        logging.log(log_level_notice, "Set log level to %s", log_level_name)
    # Return the NOTICE log level
    return log_level_notice


def get_status_language(status, detector, log_level_notice):
    """
    Get the status language, using the lingua detector if provided.
    """
    # If asked to guess...
    if detector:
        # Extract the status content
        content = status["content"]
        # Remove HTML
        content_text = html2text(content)
        # Detect language
        lingua_language = detector.detect_language_of(content_text)
        # Override
        if lingua_language == Language.ENGLISH:
            language = "en"
        elif lingua_language == Language.FRENCH:
            language = "fr"
        else:
            language = "unknown"
        logging.log(
            log_level_notice,
            "Guessing language %s for %s",
            language,
            status["url"],
        )
    else:
        # Check the status language attribute (although it may be wrong...)
        language = status["language"]
        logging.log(
            log_level_notice,
            "Taking language %s from %s",
            language,
            status["url"],
        )
    return language


def status_age_in_hours(reference, status):
    """
    Return the age of a status in hours relative to the reference datetime passed.
    """
    return int((reference - status["created_at"]).total_seconds() // 3600)


def fetch_recent_statuses(session, hashtag, hours, lazy, log_level_notice):
    """
    Query a session for statuses with hashtag, hours old at most, or less if lazy.
    """
    # Keep track of selected statuses
    fetched_statuses = []
    fetched_total = 0
    # Get the first page of the hashtag's timeline, from newer to older.
    statuses = session.timeline_hashtag(hashtag)
    # While we have statuses in the current page.
    while statuses:
        # Process next status, from newer to older.
        status = statuses.pop(0)
        # Break when reaching an old enough status
        if status_age_in_hours(datetime.now(utc), status) > hours:
            logging.log(
                log_level_notice,
                "Stopping  at %s dated %s",
                status["url"],
                status["created_at"],
            )
            break
        # Unless told to be diligent, we can stop at the first status which is already
        # favourited and reblogged.
        if lazy:
            if status["favourited"] and status["reblogged"]:
                logging.log(
                    log_level_notice,
                    "Stopping  at %s%s%s.",
                    status["url"],
                    ", favourited" if status["favourited"] else "",
                    ", reblogged" if status["favourited"] else "",
                )
                break
        # We will keep this status
        fetched_statuses.append(status)
        fetched_total += 1
        # If we have reached the last status of this page...
        if getattr(status, "_pagination_next", None):
            # ... then we refresh the status list with the next page
            # Using _pagination_next is the way to go, but Pylint complains about it, so...
            next_page = status._pagination_next  # pylint: disable=protected-access
            statuses = session.fetch_next(next_page)
    # Log our work, at NOTICE level to avoid being to noisy.
    logging.log(
        log_level_notice,
        "Fetched %d status%s",
        fetched_total,
        "es" if fetched_total != 1 else "",
    )
    # Return fetched statuses (newer first)
    return fetched_statuses


def filter_statuses_by_language(
    statuses, lang, lang_exclude, guess_lang, log_level_notice
):
    """
    Filter statuses based on language(s) required or forbidden
    """
    # If we are to guess languages, then we need lingua (but it'll use more resources)
    if guess_lang:
        languages = [Language.ENGLISH, Language.FRENCH]
        detector = LanguageDetectorBuilder.from_languages(*languages).build()
    else:
        detector = None
    # Keep track of selected statuses
    filtered_statuses = []
    # Keep count of how many statuses we considered.
    statuses_considered = 0
    statuses_filtered = 0
    # While we have statuses...
    while statuses:
        # Process next status, from newer to older.
        status = statuses.pop(0)
        # One more status considered.
        statuses_considered += 1
        # Determine status language.
        status_language = get_status_language(status, detector, log_level_notice)
        # Were any languages specified?
        if lang:
            # Do we look to restrict to, or exclude, some languages?
            if lang_exclude:
                # Looking to exclude some languages.
                if status_language in lang:
                    # Status is one of the excluded languages, so skip it.
                    continue
            else:
                # Looking to restrict to some languages
                if status_language not in lang:
                    # Status is not in any of the restricted languages, so skip it.
                    continue
        # If we've come this far, then the status should be kept
        filtered_statuses.append(status)
        statuses_filtered += 1
    # Log a tally of what we found, at notice level because we will also log actual
    # statuses processed later.
    logging.log(
        log_level_notice,
        "Filtered %d status%s by language out of %d",
        statuses_filtered,
        "es" if statuses_filtered != 1 else "",
        statuses_considered,
    )
    # Return selected statuses (newer first)
    return filtered_statuses


def status_should_be_processed(status, reverse):
    """
    Check if status should be (un)favourited or un(reblogged) or both.
    """
    should_be_processed = False
    # Extract the status attributes we will need (content is not always needed!)
    favourited = status["favourited"]
    reblogged = status["reblogged"]
    # Are we looking to reverse past actions?
    if reverse:
        # If we favourited or reblogged a status -- regardless of language!
        if favourited or reblogged:
            # Select it for unva/unreblog
            should_be_processed = True
    # Are we looking to act on recent statuses?
    else:
        # Select this status if it needs favouriting or reblogging
        if (not favourited) or (not reblogged):
            should_be_processed = True
    # Return whether status should be processed
    return should_be_processed


def eligible_statuses(statuses, reverse, log_level_notice):
    """
    Return sublist of statuses which are elegibile on lang_en and reverse criteria.
    """
    # Keep track of selected statuses
    selected_statuses = []
    # Keep count of how many statuses we considered.
    statuses_considered = 0
    statuses_selected = 0
    # While we have statuses...
    while statuses:
        # Process next status, from newer to older.
        status = statuses.pop(0)
        # One more status considered
        statuses_considered += 1
        # Should we (un)favourite or (un)reblog this status?
        if status_should_be_processed(status, reverse):
            # ... then select it
            selected_statuses.append(status)
            statuses_selected += 1
    # Log a tally of what we found, at notice level because we will also log actual
    # statuses processed later.
    logging.log(
        log_level_notice,
        "Found %d eligible status%s out of %d",
        statuses_selected,
        "es" if statuses_selected != 1 else "",
        statuses_considered,
    )
    # Return selected statuses (newer first)
    return selected_statuses


def favourite_and_reblog_status(session, status, dry_run):
    # Favourite status if it was not
    do_favourite = not status["favourited"]
    if do_favourite:
        logging.info(
            "%s %s", "Would favourite" if dry_run else "Favouriting", status["url"]
        )
        if not dry_run:
            session.status_favourite(status["id"])
    # Reblog status if it was not
    do_reblog = not status["reblogged"]
    if do_reblog:
        logging.info(
            "%s %s", "Would reblog" if dry_run else "Reblogging", status["url"]
        )
        if not dry_run:
            session.status_reblog(status["id"])
    return (do_favourite, do_reblog)


def unfavourite_and_unreblog_status(session, status, dry_run):
    # Unfavourite status if it was
    do_unfavourite = status["favorited"]
    if do_unfavourite:
        logging.info(
            "%s %s", "Would unfavourite" if dry_run else "Unfavouriting", status["url"]
        )
        if not dry_run:
            session.status_unfavourite(status["id"])
    # Unreblog status if it was
    do_unreblog = status["reblogged"]
    if do_unreblog:
        logging.info(
            "%s %s", "Would unreblog" if dry_run else "Unreblogging", status["url"]
        )
        if not dry_run:
            session.status_unreblog(status["id"])
    return (do_unfavourite, do_unreblog)


def process_single_status(session, status, reverse, dry_run):
    """
    Favourite and reblog (or the reverse) a status from a session.

    Return a pair of booleans to indicate what was done
    """
    # Are we looking to undo past actions?
    if reverse:
        # Return whether we unfavourited or unreblogged (or would have done if dry_run)
        return unfavourite_and_unreblog_status(session, status, dry_run)
    # We are not undoing a past action.
    # Return whether we favourited or reblogged (or would have done if dry_run)
    return favourite_and_reblog_status(session, status, dry_run)


def process_statuses(session, statuses, reverse, dry_run):
    """
    Favourite and reblog (or the reverse) a list of statuses from a session.
    """
    # Keep tallies of what we do.
    total_un_favourites = 0
    total_un_reblogs = 0
    statuses_processed = 0
    # As long as there are statuses left to process...
    while statuses:
        # Extract the oldest status.
        status = statuses.pop()
        # Process the status.
        (favourite, reblog) = process_single_status(session, status, reverse, dry_run)
        # We handled one more status.
        statuses_processed += 1
        if favourite:
            total_un_favourites += 1
        if reblog:
            total_un_reblogs += 1
    # Log tallies
    logging.info(
        "%s %d status%s, %d %s, %d %s.",
        "Would process" if dry_run else "Processed",
        statuses_processed,
        "es" if statuses_processed != 1 else "",
        total_un_favourites,
        "unfavourited" if reverse else "favourited",
        total_un_reblogs,
        "unreblogged" if reverse else "reblogged",
    )


def _start_session(base_url, token, username):
    """
    Create a bot, checking the actual session unsername against the provided one.
    """
    # Create the session as a Mastodon class instance
    session = Mastodon(access_token=token, api_base_url=base_url)
    # Abort if the session username does not match the specified username
    token_username = session.me()["username"]
    if username != token_username:
        raise ValueError(
            f"Username '{username}' does not match token username '{token_username}'!"
        )
    # Return the session
    return session


# Main command and its options.
@click.command()
# The base URL of the instance the bot should talk to.
@click.argument("base-url")
# The application token, and the username as a double identity check.
@click.argument("token")
@click.argument("username")
# The hashtag, hours past, language selection and language detection modes.
@click.argument("hashtag")
@click.option(
    "--hours", help="Ignore statuses older than this in hours", type=int, default=24
)
@click.option(
    "--diligent",
    is_flag=True,
    help="Do not stop at apparently already processed statuses",
)
@click.option(
    "--lang",
    help="Restrict Mastodon to some specific language(es)",
    multiple=True,
)
@click.option(
    "--lang-exclude",
    is_flag=True,
    help="Make --lang exclude, rather than include, languages",
)
@click.option(
    "--guess-lang", is_flag=True, help="Guess the status language from its text"
)
# To unfavoutite / unreblog rather than favourite / reblog.
@click.option("--reverse", is_flag=True, help="Unfav/unreglob rather than fav/reblog")
# To refrain from (un)favouriting / (un)reblogging.
@click.option("--dry-run", is_flag=True, help="Do not actually (un)fav or (un)reblog")
# To change the log level.
@click.option(
    "--log-level", help="Log level (logging module, plus NOTICE)", default="INFO"
)
# Click requires the function to have as many arguments as there are options,
# but pylint does not know that and complains, so...
def main(  # pylint: disable=too-many-arguments
    base_url,
    token,
    username,
    hashtag,
    hours,
    diligent,
    lang,
    lang_exclude,
    guess_lang,
    reverse,
    dry_run,
    log_level,
):
    """
    The main mastoboost command.
    """
    # Check contradictory options
    if lang_exclude and not lang:
        raise ValueError(
            "Option --lang-exclude is meaningless without at least one --lang option",
        )
    # Set up logging with the requested logging level
    log_level_notice = _logging_setup(log_level)
    # Start the Mastodon session
    session = _start_session(base_url, token, username)
    # Fetch the statuses we might process
    statuses = fetch_recent_statuses(
        session, hashtag, hours, diligent and not reverse, log_level_notice
    )
    # Filter status by language if requested
    if lang and not reverse:
        statuses = filter_statuses_by_language(
            statuses, lang, lang_exclude, guess_lang, log_level_notice
        )
    # Keep only the eligible ones
    statuses = eligible_statuses(statuses, reverse, log_level_notice)
    # Process eligible statuses unless told not to.
    process_statuses(session, statuses, reverse, dry_run)


if __name__ == "__main__":
    # Click provides the parameters, but pylint does not know that, so...
    main()  # pylint: disable=no-value-for-parameter
